/*
  This program transforms two DLV style programs to two first-order theories.
  If both are unsatisfiable, the programs are strongly equivalent.
  Copyright (C) 2004 Patrick Traxler

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "st.h"

extern int yyparse();
extern FILE* yyin;

/*
    Generates Sigma. Please see the work out for
    further information.

    Uses the global variable st_RS.
*/
string genSigma() {
    string s, t, u = "";
    for (s = st_RS.iterate(0); s != ""; s = st_RS.iterate(1)) {
        t = "";
        for (int i = 0; i < st_RS.getArity(s); i++) {
            char c[sizeof(int)];
            sprintf(c, "%d", i + 1);
            t+= "X";
            t+= c;
            t+= (i == st_RS.getArity(s)-1) ? "" : ",";
        }
        if (st_RS.getArity(s) == 0) {
            t = s + "_:- " + s + ".\n";
        }
        else {
            t = s + "_(" + t + "):- " + s + "(" + t + ").\n";
        }
        u = t + u;
    }
    return u;
}

/*
    Generates U. Please see the work out for
    further information.

    Uses the global variable st_CS.
*/
string genU() {
    string s, t, u = "";
    int j;
    for (s = st_CS.iterate(0), j = st_CS.size()-1; s != ""; s = st_CS.iterate(1), j--) {
        t = "";
        for (int i = 0; i < st_CS.size(); i++) {
            char c[sizeof(int)];
            sprintf(c, "%d", i + 1);
            t+= (i == j) ? "" : "-";
            t+= "u_";
            t+= c;
            t+= "(" + s + "). ";
        }
        u = t + "\n" + u;
    }
    return u;
}

/*
    Generates Gamma. Please see the work out for
    further information.
*/
string genGamma(string prog) {
    Rule r;
    string lit,
           o1 = "",
           o2 = "";

    while (r.nextRule(prog)) {
        lit = r.nextLiteral();
        if (lit == ":-") {
            o1 += "false:- ";
            o2 += "false:- ";
            lit = r.nextLiteral();
        }

        do
        {
            //Argumentlist:
            int    pos = lit.find("(", 0);
            string arg = "";
            if (pos != string::npos) {
                arg = lit.substr(pos);
                lit.erase(pos);
            }
            //o1:
            if (lit[0] == '~') {
                if (lit[1] == '-')
                    o1 += lit.substr(2) + "_";
                else
                    o1 += "-" + lit.substr(1) + "_";
            }
            else {
                o1 += lit;
            }
            //o2:
            if (lit[0] == '~') {
                if (lit[1] == '-')
                    o2 += lit.substr(2) + "_";
                else
                    o2 += "-" + lit.substr(1) + "_";
            }
            else {
                o2 += lit + "_";
            }
            o1 += arg;
            o2 += arg;
            //Next literal:
            lit = r.nextLiteral();
            if (lit == ":-") {
                o1+= ":- ";
                o2+= ":- ";
                lit  = r.nextLiteral();
                continue;
            }
            else if (lit == "")
                break;
            else {
                o1 += ",";
                o2 += ",";
            }
        }
        while (1);
        o1+= ".\n";
        o2+= ".\n";
    }
    return o1 + "\n" + o2;
}

/*
    Generates Gamma*. Please see the work out for
    further information.
*/
string genGammaStar(string prog) {
    Rule r, s;
    string c1, c2,
           o3 = "",
           o4 = "",
           o5 = "",
           l_r, l_s,
           a_r, a_s;
    bool neg;

    while (r.nextRule(prog)) {
        s = r;
        r.skolemize();
        s.skolemize();

        c1  = st_CS.newSymbol("c");
        c2  = st_CS.newSymbol("c");
        neg = true;
        while ((l_r = r.nextLiteral()) != "") {
            l_s = s.nextLiteral();
            if (l_r == ":-") {
                neg = false;
                l_r = r.nextLiteral();
                l_s = s.nextLiteral();
            }
            //Argumentlist:
            int pos    = l_r.find("(", 0);
            string a_r = "";
            if (pos != string::npos) {
                a_r = l_r.substr(pos);
                l_r.erase(pos);
            }
            pos = l_s.find("(", 0);
            string a_s = "";
            if (pos != string::npos) {
                a_s = l_s.substr(pos);
                l_s.erase(pos);
            }
            //o3:
            if (l_r[0] == '~') {
                if (l_r[1] == '-')
                    o3 += l_r.substr(2) + "_";
                else
                    o3+= "-" + l_r.substr(1) + "_";
            }
            else {
                if (l_r[0] == '-')
                    o3 += (neg ? "" : "-") + l_r.substr(1);
                else
                    o3 += (neg ? "-" : "") + l_r;
            }
            o3 += a_r + ":- s__(" + c1 + ").\n";
            //o4:
            if (l_s[0] == '~') {
                if (l_s[1] == '-')
                    o4 += l_s.substr(2) + "_";
                else
                    o4 += "-" + l_s.substr(1) + "_";
            }
            else {
                if (l_s[0] == '-')
                    o4 += (neg ? "" : "-") + l_s.substr(1) + "_";
                else
                    o4 += (neg ? "-" : "") + l_s + "_";
            }
            o4 += a_s + ":- s__(" + c2 + ").\n";
        }
        o5+= "s__(" + c1 + ")," + "s__(" + c2 + "),";
    }
    o5.erase(o5.length()-1, 1);
    o5+= ".";
    return o3 + "\n" + o4 + "\n" + o5;
}

/*
    Two DLV programs are read in. They are strongly equivalent
    iff Darwin finds a refutation in "out.1.tme" and in "out.2.tme"
    which are written out.
*/
int main(int argn, char** argv) {
    string prog1 = "",
           prog2 = "";
    char b[1024];

    if (argn < 3) {
        cout << "Command: dlv2darwin filename1 filename2" << endl;
        return 3;
    }

    yyin = fopen(argv[1], "r");
    if (yyin == NULL) {
        cout << "Could not find file " << argv[1] << endl;
        return 3;
    }
    yyparse();
    rewind(yyin);
    while (fgets(b, 1028, yyin)) {
        prog1+= b;
    }
    fclose(yyin);

    yyin = fopen(argv[2], "r");
    if (yyin == NULL) {
        cout << "Could not find file " << argv[2] <<  endl;
        return 3;
    }
    yyparse();
    rewind(yyin);
    while (fgets(b, 1028, yyin)) {
        prog2+= b;
    }
    fclose(yyin);

    ofstream f1("out.1.tme");
    ofstream f2("out.2.tme");

    f1 << "%Sigma:\n" << genSigma() << endl;
    f1 << "%Unique Name Axioms:\n" << genU() << endl;
    f1 << "%Gamma(Pi):\n" << genGamma(prog1) << endl;
    f1 << "%GammaStar(Pi'):\n" << genGammaStar(prog2) << endl;

    f2 << "%Sigma:\n" << genSigma() << endl;
    f2 << "%Unique Name Axioms:\n" << genU() << endl;
    f2 << "%Gamma(Pi'):\n" << genGamma(prog2) << endl;
    f2 << "%GammaStar(Pi):\n" << genGammaStar(prog1) << endl;

    return 0;
}
